"""Contains modules for controlling the PSLab's built-in instruments.

The built-in instruments are ``LogicAnalyzer``, ``Multimeter``,
``Oscilloscope``, ``PowerSupply``, ``PWMGenerator``, and ``WaveformGenerator``.
"""
